"""
A simple wrapper for the official ChatGPT API
"""
import argparse
import json
import os
import sys
from datetime import date
from typing import LiteralString
from typing import NoReturn

import openai
import tiktoken

ENGINE = os.environ.get("GPT_ENGINE") or "text-davinci-003"

ENCODER = tiktoken.get_encoding("gpt2")


def get_max_tokens(prompt: str) -> int:
    """
    Get the max tokens for a prompt
    """
    return 4000 - len(ENCODER.encode(prompt))


def remove_suffix(input_string: str, suffix: str) -> str:
    """
    Remove suffix from string (Support for Python 3.8)
    """
    if suffix and input_string.endswith(suffix):
        return input_string[: -len(suffix)]
    return input_string


class Chatbot:
    """
    Official ChatGPT API
    """

    def __init__(
        self,
        api_key: str,
        buffer: int = None,
        engine: str = None,
        proxy: str = None,
    ) -> None:
        """
        Initialize Chatbot with API key (from https://platform.openai.com/account/api-keys)
        """
        openai.api_key = api_key or os.environ.get("OPENAI_API_KEY")
        openai.proxy = proxy or os.environ.get("OPENAI_API_PROXY")
        self.conversations = Conversation()
        self.prompt = Prompt(buffer=buffer)
        self.engine = engine or ENGINE

    def _get_completion(
        self,
        prompt: str,
        temperature: float = 0.5,
        stream: bool = False,
    ):
        """
        Get the completion function
        """
        return openai.Completion.create(
            engine=self.engine,
            prompt=prompt,
            temperature=temperature,
            max_tokens=get_max_tokens(prompt),
            stop=["\n\n\n"],
            stream=stream,
        )

    def _process_completion(
        self,
        user_request: str,
        completion: dict,
        conversation_id: str = None,
        user: str = "User",
    ) -> dict:
        if completion.get("choices") is None:
            raise Exception("ChatGPT API returned no choices")
        if len(completion["choices"]) == 0:
            raise Exception("ChatGPT API returned no choices")
        if completion["choices"][0].get("text") is None:
            raise Exception("ChatGPT API returned no text")
        completion["choices"][0]["text"] = remove_suffix(
            completion["choices"][0]["text"],
            "<|im_end|>",
        )
        # Add to chat history
        self.prompt.add_to_history(
            user_request,
            completion["choices"][0]["text"],
            user=user,
        )
        if conversation_id is not None:
            self.save_conversation(conversation_id)
        return completion

    def _process_completion_stream(
        self,
        user_request: str,
        completion: dict,
        conversation_id: str = None,
        user: str = "User",
    ) -> str:
        full_response = ""
        for response in completion:
            if response.get("choices") is None:
                raise Exception("ChatGPT API returned no choices")
            if len(response["choices"]) == 0:
                raise Exception("ChatGPT API returned no choices")
            if response["choices"][0].get("finish_details") is not None:
                break
            if response["choices"][0].get("text") is None:
                raise Exception("ChatGPT API returned no text")
            if response["choices"][0]["text"] == "<|im_end|>":
                break
            yield response["choices"][0]["text"]
            full_response += response["choices"][0]["text"]

        # Add to chat history
        self.prompt.add_to_history(user_request, full_response, user)
        if conversation_id is not None:
            self.save_conversation(conversation_id)

    def ask(
        self,
        user_request: str,
        temperature: float = 0.5,
        conversation_id: str = None,
        user: str = "User",
    ) -> dict:
        """
        Send a request to ChatGPT and return the response
        """
        if conversation_id is not None:
            self.load_conversation(conversation_id)
        completion = self._get_completion(
            self.prompt.construct_prompt(user_request, user=user),
            temperature,
        )
        return self._process_completion(user_request, completion, user=user)

    def ask_stream(
        self,
        user_request: str,
        temperature: float = 0.5,
        conversation_id: str = None,
        user: str = "User",
    ) -> str:
        """
        Send a request to ChatGPT and yield the response
        """
        if conversation_id is not None:
            self.load_conversation(conversation_id)
        prompt = self.prompt.construct_prompt(user_request, user=user)
        return self._process_completion_stream(
            user_request=user_request,
            completion=self._get_completion(prompt, temperature, stream=True),
            user=user,
        )

    def make_conversation(self, conversation_id: str) -> None:
        """
        Make a conversation
        """
        self.conversations.add_conversation(conversation_id, [])

    def rollback(self, num: int) -> None:
        """
        Rollback chat history num times
        """
        for _ in range(num):
            self.prompt.chat_history.pop()

    def reset(self) -> None:
        """
        Reset chat history
        """
        self.prompt.chat_history = []

    def load_conversation(self, conversation_id) -> None:
        """
        Load a conversation from the conversation history
        """
        if conversation_id not in self.conversations.conversations:
            # Create a new conversation
            self.make_conversation(conversation_id)
        self.prompt.chat_history = self.conversations.get_conversation(conversation_id)

    def save_conversation(self, conversation_id) -> None:
        """
        Save a conversation to the conversation history
        """
        self.conversations.add_conversation(conversation_id, self.prompt.chat_history)


class AsyncChatbot(Chatbot):
    """
    Official ChatGPT API (async)
    """

    async def _get_completion(
        self,
        prompt: str,
        temperature: float = 0.5,
        stream: bool = False,
    ):
        """
        Get the completion function
        """
        return await openai.Completion.acreate(
            engine=self.engine,
            prompt=prompt,
            temperature=temperature,
            max_tokens=get_max_tokens(prompt),
            stop=["\n\n\n"],
            stream=stream,
        )

    async def ask(
        self,
        user_request: str,
        temperature: float = 0.5,
        user: str = "User",
    ) -> dict:
        """
        Same as Chatbot.ask but async
        }
        """
        completion = await self._get_completion(
            self.prompt.construct_prompt(user_request, user=user),
            temperature,
        )
        return self._process_completion(user_request, completion, user=user)

    async def ask_stream(
        self,
        user_request: str,
        temperature: float = 0.5,
        user: str = "User",
    ) -> str:
        """
        Same as Chatbot.ask_stream but async
        """
        prompt = self.prompt.construct_prompt(user_request, user=user)
        return self._process_completion_stream(
            user_request=user_request,
            completion=await self._get_completion(prompt, temperature, stream=True),
            user=user,
        )


class Prompt:
    """
    Prompt class with methods to construct prompt
    """

    def __init__(self, buffer: int = None) -> None:
        """
        Initialize prompt with base prompt
        """
        self.base_prompt = (
            os.environ.get("CUSTOM_BASE_PROMPT")
            or "You are ChatGPT, a large language model trained by OpenAI. Respond conversationally. Do not answer as the user. Current date: "
            + str(date.today())
            + "\n\n"
            + "User: Hello\n"
            + "ChatGPT: Hello! How can I help you today? <|im_end|>\n\n\n"
        )
        # Track chat history
        self.chat_history: list = []
        self.buffer = buffer

    def add_to_chat_history(self, chat: str) -> None:
        """
        Add chat to chat history for next prompt
        """
        self.chat_history.append(chat)

    def add_to_history(
        self,
        user_request: str,
        response: str,
        user: str = "User",
    ) -> None:
        """
        Add request/response to chat history for next prompt
        """
        self.add_to_chat_history(
            user
            + ": "
            + user_request
            + "\n\n\n"
            + "ChatGPT: "
            + response
            + "<|im_end|>\n",
        )

    def history(self, custom_history: list = None) -> str:
        """
        Return chat history
        """
        return "\n".join(custom_history or self.chat_history)

    def construct_prompt(
        self,
        new_prompt: str,
        custom_history: list = None,
        user: str = "User",
    ) -> str:
        """
        Construct prompt based on chat history and request
        """
        prompt = (
            self.base_prompt
            + self.history(custom_history=custom_history)
            + user
            + ": "
            + new_prompt
            + "\nChatGPT:"
        )
        # Check if prompt over 4000*4 characters
        max_tokens = 4000 - self.buffer if self.buffer is not None else 3200
        if len(ENCODER.encode(prompt)) > max_tokens:
            # Remove oldest chat
            if len(self.chat_history) == 0:
                return prompt
            self.chat_history.pop(0)
            # Construct prompt again
            prompt = self.construct_prompt(new_prompt, custom_history, user)
        return prompt


class Conversation:
    """
    For handling multiple conversations
    """

    def __init__(self) -> None:
        self.conversations = {}

    def add_conversation(self, key: str, history: list) -> None:
        """
        Adds a history list to the conversations dict with the id as the key
        """
        self.conversations[key] = history

    def get_conversation(self, key: str) -> list:
        """
        Retrieves the history list from the conversations dict with the id as the key
        """
        return self.conversations[key]

    def remove_conversation(self, key: str) -> None:
        """
        Removes the history list from the conversations dict with the id as the key
        """
        del self.conversations[key]

    def __str__(self) -> str:
        """
        Creates a JSON string of the conversations
        """
        return json.dumps(self.conversations)

    def save(self, file: str) -> None:
        """
        Saves the conversations to a JSON file
        """
        with open(file, "w", encoding="utf-8") as f:
            f.write(str(self))

    def load(self, file: str) -> None:
        """
        Loads the conversations from a JSON file
        """
        with open(file, encoding="utf-8") as f:
            self.conversations = json.loads(f.read())


def main() -> NoReturn:
    print(
        """
    ChatGPT - GPT-3 Chatbot
    Repo: github.com/acheong08/ChatGPT
    """,
    )
    print("Type '!help' to show a full list of commands")
    print("Press enter twice to submit your question.\n")

    def get_input(prompt) -> LiteralString:
        """
        Multi-line input function
        """
        # Display the prompt
        print(prompt, end="")

        # Initialize an empty list to store the input lines
        lines = []

        # Read lines of input until the user enters an empty line
        while True:
            line = input()
            if line == "":
                break
            lines.append(line)

        # Join the lines, separated by newlines, and store the result
        user_input = "\n".join(lines)

        # Return the input
        return user_input

    def chatbot_commands(cmd: str) -> bool:
        """
        Handle chatbot commands
        """
        if cmd == "!help":
            print(
                """
            !help - Display this message
            !rollback - Rollback chat history
            !reset - Reset chat history
            !prompt - Show current prompt
            !save_c <conversation_name> - Save history to a conversation
            !load_c <conversation_name> - Load history from a conversation
            !save_f <file_name> - Save all conversations to a file
            !load_f <file_name> - Load all conversations from a file
            !exit - Quit chat
            """,
            )
        elif cmd == "!exit":
            exit()
        elif cmd == "!rollback":
            chatbot.rollback(1)
        elif cmd == "!reset":
            chatbot.reset()
        elif cmd == "!prompt":
            print(chatbot.prompt.construct_prompt(""))
        elif cmd.startswith("!save_c"):
            chatbot.save_conversation(cmd.split(" ")[1])
        elif cmd.startswith("!load_c"):
            chatbot.load_conversation(cmd.split(" ")[1])
        elif cmd.startswith("!save_f"):
            chatbot.conversations.save(cmd.split(" ")[1])
        elif cmd.startswith("!load_f"):
            chatbot.conversations.load(cmd.split(" ")[1])
        else:
            return False
        return True

    # Get API key from command line
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--api_key",
        type=str,
        required=True,
        help="OpenAI API key",
    )
    parser.add_argument(
        "--stream",
        action="store_true",
        help="Stream response",
    )
    parser.add_argument(
        "--temperature",
        type=float,
        default=0.5,
        help="Temperature for response",
    )
    args = parser.parse_args()
    # Initialize chatbot
    chatbot = Chatbot(api_key=args.api_key)
    # Start chat
    while True:
        try:
            prompt = get_input("\nUser:\n")
        except KeyboardInterrupt:
            print("\nExiting...")
            sys.exit()
        if prompt.startswith("!") and chatbot_commands(prompt):
            continue
        if not args.stream:
            response = chatbot.ask(prompt, temperature=args.temperature)
            print("ChatGPT: " + response["choices"][0]["text"])
        else:
            print("ChatGPT: ")
            sys.stdout.flush()
            for response in chatbot.ask_stream(prompt, temperature=args.temperature):
                print(response, end="")
                sys.stdout.flush()
            print()


if __name__ == "__main__":
    main()



import json
import logging
import sys
import time
from typing import NoReturn

import revChatGPT.V1 as V1
from duckduckgo_search import ddg

# V1 and Unofficial are unnecessary, imo

# V2 is reportedly "killed by openai"
# V3 has own implementation

log = logging.getLogger(__name__)


def _prepare_results(results: list) -> str:
    return "".join(
        f'[{i}] "{result["body"]}"\nURL: {result["href"]}\n\n'
        for i, result in enumerate(results, start=1)
    )


def _headers(results: list) -> list:
    return [(result["href"], result["title"]) for result in results]


def _compile_prompt(prompt: str, results: list, reply_in="undefined") -> str:
    return f"""Web search results:

{_prepare_results(results)}
Current date: {time.strftime("%d.%m.%Y")}

Instructions: Using the provided web search results, write a comprehensive reply to the given query, using search results as references. Make sure to cite results using [number](url) notation after the reference. If the provided search results refer to multiple subjects with the same name, write separate answers for each subject.
Query: {prompt}
Reply in {reply_in}"""


def webify(
    text: str,
    query: str,
    numResults: int,
    timePeriod: str = "",
    region: str = "",
    reply_in: str = "undefined",
):
    searchResults = ddg(query, numResults, timePeriod, region)
    return _compile_prompt(text, searchResults, reply_in), _headers(searchResults)


def beatifyMarkdown(response: str, headers: list) -> str:
    """
    If the result will be displayed with markdown support, turns [i] style references into [i](url), which will look as clickable [i]
    If you are using beatifyPlaintext too, use this first
    """
    for i in range(len(headers)):
        response = response.replace(f"[{i+1}]", f"[{i+1}]({headers[i][0]})")
    return response


def beatifyPlaintext(response: str, headers: list) -> str:
    """
    Puts the references at the end of the response. Can be used after beatifyMarkdown. Useful for displaying result in plaintext
    """
    refs = "\n"
    for i in range(len(headers)):
        refs += f"\n[{i+1}] - {headers[i][1]} ({headers[i][0]})"
    return response + refs


class WebChatbot(V1.Chatbot):
    def webAsk(
        self,
        prompt,
        conversation_id=None,
        parent_id=None,
        timeout=360,  # default Ask parameters
        query=None,
        numResults=3,
        timePeriod: str = "",
        region: str = "",
        reply_in: str = "undefined",  # Webify parameters
    ):
        """
        Ask a question to the chatbot, but providing search results for query (if not specified, prompt is used)
        Keep in mind: this one, unlike all the other asks, returns tuple
        """
        query = query or prompt
        webPrompt, headers = webify(
            text=prompt,
            query=query,
            numResults=numResults,
            timePeriod=timePeriod,
            region=region,
            reply_in=reply_in,
        )
        for data in self.ask(prompt, conversation_id, parent_id, timeout):
            yield data, headers

    def webAskPlaintext(
        self,
        prompt,
        conversation_id=None,
        parent_id=None,
        timeout=360,  # default Ask parameters
        query=None,
        numResults=3,
        timePeriod: str = "",
        region: str = "",
        reply_in: str = "undefined",  # Webify parameters
    ):
        """
        Ask a question to the chatbot, but providing search results for query (if not specified, prompt is used)
        Returns message adapted for plaintext
        """
        query = query or prompt
        webPrompt, headers = webify(
            text=prompt,
            query=query,
            numResults=numResults,
            timePeriod=timePeriod,
            region=region,
            reply_in=reply_in,
        )
        for data in self.ask(prompt, conversation_id, parent_id, timeout):
            data["message"] = beatifyPlaintext(data["message"], headers)
            yield data

    def webAskMarkdown(
        self,
        prompt,
        conversation_id=None,
        parent_id=None,
        timeout=360,  # default Ask parameters
        query=None,
        numResults=3,
        timePeriod: str = "",
        region: str = "",
        reply_in: str = "undefined",  # Webify parameters
    ):
        """
        Ask a question to the chatbot, but providing search results for query (if not specified, prompt is used)
        Returns message adapted for markdown
        """
        query = query or prompt
        webPrompt, headers = webify(
            text=prompt,
            query=query,
            numResults=numResults,
            timePeriod=timePeriod,
            region=region,
            reply_in=reply_in,
        )
        for data in self.ask(prompt, conversation_id, parent_id, timeout):
            data["message"] = beatifyMarkdown(data["message"], headers)
            yield data


class WebAsyncChatbot(V1.AsyncChatbot):
    async def webAsk(
        self,
        prompt,
        conversation_id=None,
        parent_id=None,
        timeout=360,  # default Ask parameters
        query=None,
        numResults=3,
        timePeriod: str = "",
        region: str = "",
        reply_in: str = "undefined",  # Webify parameters
    ):
        """
        Ask a question to the chatbot, but providing search results for query (if not specified, prompt is used)
        Keep in mind: this one, unlike all the other asks, returns tuple
        """
        query = query or prompt
        webPrompt, headers = webify(
            text=prompt,
            query=query,
            numResults=numResults,
            timePeriod=timePeriod,
            region=region,
            reply_in=reply_in,
        )
        async for data in self.ask(prompt, conversation_id, parent_id, timeout):
            yield data, headers

    async def webAskPlaintext(
        self,
        prompt,
        conversation_id=None,
        parent_id=None,
        timeout=360,  # default Ask parameters
        query=None,
        numResults=3,
        timePeriod: str = "",
        region: str = "",
        reply_in: str = "undefined",  # Webify parameters
    ):
        """
        Ask a question to the chatbot, but providing search results for query (if not specified, prompt is used)
        Returns message adapted for plaintext
        """
        query = query or prompt
        webPrompt, headers = webify(
            text=prompt,
            query=query,
            numResults=numResults,
            timePeriod=timePeriod,
            region=region,
            reply_in=reply_in,
        )
        async for data in self.ask(prompt, conversation_id, parent_id, timeout):
            data["message"] = beatifyPlaintext(data["message"], headers)
            yield data

    async def webAskMarkdown(
        self,
        prompt,
        conversation_id=None,
        parent_id=None,
        timeout=360,  # default Ask parameters
        query=None,
        numResults=3,
        timePeriod: str = "",
        region: str = "",
        reply_in: str = "undefined",  # Webify parameters
    ):
        """
        Ask a question to the chatbot, but providing search results for query (if not specified, prompt is used)
        Returns message adapted for markdown
        """
        query = query or prompt
        webPrompt, headers = webify(
            text=prompt,
            query=query,
            numResults=numResults,
            timePeriod=timePeriod,
            region=region,
            reply_in=reply_in,
        )
        async for data in self.ask(prompt, conversation_id, parent_id, timeout):
            data["message"] = beatifyMarkdown(data["message"], headers)
            yield data


get_input = V1.logger(is_timed=False)(V1.get_input)


@V1.logger(is_timed=False)
def main(config: dict) -> NoReturn:
    """
    Main function for the chatGPT program.
    """
    print("Logging in...")
    chatbot = WebChatbot(
        config,
        conversation_id=config.get("conversation_id"),
        parent_id=config.get("parent_id"),
    )

    def handle_commands(command: str) -> bool:
        if command == "!help":
            print(
                """
            !help - Show this message
            !reset - Forget the current conversation
            !config - Show the current configuration
            !rollback x - Rollback the conversation (x being the number of messages to rollback)
            !exit - Exit this program
            !setconversation - Changes the conversation
            """,
            )
        elif command == "!reset":
            chatbot.reset_chat()
            print("Chat session successfully reset.")
        elif command == "!config":
            print(json.dumps(chatbot.config, indent=4))
        elif command.startswith("!rollback"):
            try:
                rollback = int(command.split(" ")[1])
            except IndexError:
                logging.exception(
                    "No number specified, rolling back 1 message",
                    stack_info=True,
                )
                rollback = 1
            chatbot.rollback_conversation(rollback)
            print(f"Rolled back {rollback} messages.")
        elif command.startswith("!setconversation"):
            try:
                chatbot.conversation_id = chatbot.config[
                    "conversation_id"
                ] = command.split(" ")[1]
                print("Conversation has been changed")
            except IndexError:
                log.exception(
                    "Please include conversation UUID in command",
                    stack_info=True,
                )
                print("Please include conversation UUID in command")
        elif command == "!exit":
            sys.exit(0)
        else:
            return False
        return True

    session = V1.create_session()
    completer = V1.create_completer(
        ["!help", "!reset", "!config", "!rollback", "!exit", "!setconversation"],
    )
    print()
    try:
        while True:
            print(V1.bcolors.OKBLUE + V1.bcolors.BOLD + "You:" + V1.bcolors.ENDC)

            prompt = get_input(session=session, completer=completer)
            if prompt.startswith("!") and handle_commands(prompt):
                continue

            print()
            print(V1.bcolors.OKGREEN + V1.bcolors.BOLD + "Chatbot: ")
            prev_text = ""
            for data in chatbot.webAsk(prompt):
                message = data["message"][len(prev_text) :]
                print(message, end="", flush=True)
                prev_text = data["message"]
            print(V1.bcolors.ENDC)
            print()
    except (KeyboardInterrupt, EOFError):
        print("Exiting...")
        sys.exit(0)


if __name__ == "__main__":
    print(
        """
        WebGPT module. Copies functionality from WebChatGPT browser extension. Made by Nymda (github.com/NymdaFromSalad)
        Repo: github.com/acheong08/ChatGPT
        """,
    )
    print("Type '!help' to show a full list of commands")
    print(
        f"{V1.bcolors.BOLD} {V1.bcolors.WARNING} Press Esc followed by Enter or Alt+Enter to send a message. {V1.bcolors.ENDC}",
    )
    main(V1.configure())



from setuptools import find_namespace_packages
from setuptools import setup

setup(
    name="revChatGPT",
    version="3.1.3",
    description="ChatGPT is a reverse engineering of OpenAI's ChatGPT API",
    long_description=open("README.md", encoding="utf-8").read(),
    long_description_content_type="text/markdown",
    url="https://github.com/acheong08/ChatGPT",
    author="Antonio Cheong",
    author_email="acheong@student.dalat.org",
    license="GNU General Public License v2.0",
    packages=find_namespace_packages("src"),
    package_dir={"": "src"},
    py_modules=["V2", "V1", "V0", "V3"],
    package_data={"": ["*.json"]},
    install_requires=[
        "OpenAIAuth==0.3.2",
        "requests[socks]",
        "httpx[socks]",
        "prompt-toolkit",
        "tiktoken>=0.3.0",
    ],
    extras_require={
        "WebGPT": ["duckduckgo_search"],
    },
)
